import inspect
import random

from appworld.common.datetime import DateTime, Time, _DateTime
from appworld.common.types import Number
from appworld.common.utils import TItem, range_plus


def set_random_seed(
    seed: int | None = None,
    python: bool = True,
    faker_: bool = False,
    model_factory: bool = False,
) -> None:
    if seed is None:
        return
    if python:
        random.seed(seed)
    # NOTE: numpy is not a dependency, so dropped setting its seed.
    if faker_:
        from faker import Faker

        Faker.seed(seed)
    if model_factory:
        from appworld.apps.model_lib import ModelFactory

        ModelFactory.seed_random(seed)


def choose_from_range(
    start_at: Number | _DateTime | Time, end_at: Number | _DateTime | Time
) -> TItem:
    if isinstance(start_at, Number) and isinstance(end_at, Number):
        return choose_from_number_range(start_at, end_at)
    elif isinstance(start_at, _DateTime) and isinstance(end_at, _DateTime):
        return choose_from_datetime_range(start_at, end_at)
    elif isinstance(start_at, Time) and isinstance(end_at, Time):
        return choose_from_time_range(start_at, end_at)
    raise ValueError(
        "The start_at and end_at must be of the same type: Number, DateTime, or Time. "
        f"The start_at is of type {type(start_at)} and the end_at is of type {type(end_at)}."
    )


def sample_from_range(
    count: int,
    start_at: Number | _DateTime | Time,
    end_at: Number | _DateTime | Time,
) -> list[TItem]:
    if isinstance(start_at, Number) and isinstance(end_at, Number):
        return sample_from_number_range(count, start_at, end_at)
    elif isinstance(start_at, _DateTime) and isinstance(end_at, _DateTime):
        return sample_from_datetime_range(count, start_at, end_at)
    elif isinstance(start_at, Time) and isinstance(end_at, Time):
        return sample_from_time_range(count, start_at, end_at)
    raise ValueError(
        "The start_at and end_at must be of the same type: Number, DateTime, or Time. "
        f"The start_at is of type {type(start_at)} and the end_at is of type {type(end_at)}."
    )


def choose_from_datetime_range(
    start_datetime: _DateTime | None = None, end_datetime: _DateTime | None = None
) -> _DateTime:
    if start_datetime is None:
        start_datetime = DateTime(2010, 1, 1)
    if end_datetime is None:
        end_datetime = DateTime.now()
    assert start_datetime < end_datetime, "The start_datetime must be before the end_datetime."
    delta_seconds = (end_datetime - start_datetime).in_seconds()
    random_seconds = random.randint(0, delta_seconds)
    return start_datetime.add(seconds=random_seconds)


def sample_from_datetime_range(
    count: int, start_datetime: _DateTime | None = None, end_datetime: _DateTime | None = None
) -> list[_DateTime]:
    return [choose_from_datetime_range(start_datetime, end_datetime) for _ in range(count)]


def choose_from_time_range(
    start_time: Time = Time(0, 0, 0), end_time: Time = Time(23, 59, 59)
) -> Time:
    assert start_time < end_time, "The start_time must be before the end_time."
    delta_seconds = (end_time - start_time).in_seconds()
    random_seconds = random.randint(0, delta_seconds)
    return start_time.add(seconds=random_seconds)


def sample_from_time_range(
    count: int, start_time: Time = Time(0, 0, 0), end_time: Time = Time(23, 59, 59)
) -> list[Time]:
    return [choose_from_time_range(start_time, end_time) for _ in range(count)]


def choose_from_list(
    items: list[TItem],
    exclude_items: list[TItem] | None = None,
    remove_sampled: bool = False,
    weights: list[float] | None = None,
) -> TItem | None:
    dict_keys = type({}.keys())
    dict_values = type({}.values())
    is_generator = inspect.isgenerator(items)
    if isinstance(items, (dict_keys, dict_values)) or is_generator:
        items = list(items)
    if isinstance(items, set):
        if not exclude_items and not weights and not remove_sampled:
            items = list(items)
        else:
            raise ValueError(
                "Cannot choose from a set with exclude_items, weights, or remove_sampled."
            )
    original_items = items
    exclude_items = exclude_items or []
    weights = [1.0] * len(items) if weights is None else list(weights)
    if len(weights) != len(items):
        raise ValueError("Length of weights must be equal to the length of items")
    indices = [index for index, item in enumerate(items) if item not in exclude_items]
    items = [items[index] for index in indices]
    weights = [weights[index] for index in indices]
    chosen_item = random.choices(items, weights)[0]
    if remove_sampled:
        original_items.remove(chosen_item)
    return chosen_item


def sample_from_list(
    count: int,
    items: list[TItem],
    exclude_items: list[TItem] | None = None,
    allow_fewer: bool = False,
    remove_sampled: bool = False,
    weights: list[float] | None = None,
) -> list[TItem]:
    dict_keys = type({}.keys())
    dict_values = type({}.values())
    is_generator = inspect.isgenerator(items)
    if isinstance(items, (dict_keys, dict_values)) or is_generator:
        items = list(items)
    if isinstance(items, set):
        if not exclude_items and not weights and not remove_sampled:
            items = list(items)
        else:
            raise ValueError(
                "Cannot sample from a set with exclude_items, weights, or remove_sampled."
            )
    original_items = items
    exclude_items = exclude_items or []
    weights = [1.0] * len(items) if weights is None else list(weights)
    if len(weights) != len(items):
        raise ValueError("Length of weights must be equal to the length of items")
    indices = [index for index, item in enumerate(items) if item not in exclude_items]
    items = [items[index] for index in indices]
    weights = [weights[index] for index in indices]
    if allow_fewer:
        count = min(count, len(items))
    if count == 0:
        return []
    selected_items = []
    for _ in range(count):
        if not items:
            break
        indices = list(range(len(items)))
        if allow_fewer and sum(weights) == 0:
            continue
        chosen_index = random.choices(indices, weights)[0]
        selected_items.append(items[chosen_index])
        items.pop(chosen_index)
        weights.pop(chosen_index)
    if remove_sampled:
        for item in selected_items:
            original_items.remove(item)
    if not allow_fewer and len(selected_items) < count:
        raise ValueError(
            "Could not sample enough items. "
            "Try setting allow_fewer to True or increasing the number of items."
        )
    return selected_items


def choose_from_number_range(start_number: Number, end_number: Number) -> list[Number]:
    if isinstance(start_number, int) and isinstance(end_number, int):
        return random.randint(start_number, end_number)
    else:
        return choose_from_list(range_plus(start_number, end_number))


def sample_from_number_range(count: int, start_number: Number, end_number: Number) -> Number:
    if isinstance(start_number, int) and isinstance(end_number, int):
        return [random.randint(start_number, end_number) for _ in range(count)]
    else:
        return sample_from_list(count, range_plus(start_number, end_number))


def is_true(probability: float) -> bool:
    assert 0 <= probability <= 1, "Not a valid probability."
    return random.random() <= probability
